"""Provides SQLAlchemy based model classes and utility functions which form the
base on which to implement a simple role based security for an applicatin.
"""
import bcrypt
from sqlalchemy import (
    Column, Boolean, DateTime, ForeignKey, Integer,
    String, Table
)
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func
from elixr import utils

from ..mixins import IdMixin
from ..meta import Model



## FUNCS
def generate_confirmation_hash():
    return utils.generate_random_digest(num_bytes=14)


def _check_password(passwd, passwd_hash):
    expected_hash = passwd_hash.encode('utf8')
    return bcrypt.checkpw(passwd.encode('utf8'), expected_hash)


def _hash_password(passwd):
    pwhash = bcrypt.hashpw(passwd.encode('utf8'), bcrypt.gensalt())
    return pwhash.decode('utf8')


## MODELS
# many-to-many relation between users and roles
auth_users_roles = Table(
    'auth_users_roles',
    Model.metadata,
    Column('user_id', Integer, ForeignKey('auth_users.id')),
    Column('role_id', Integer, ForeignKey('auth_roles.id'))
)


class User(Model, IdMixin):
    """A model for storing User data.
    """
    __tablename__ = 'auth_users'
    
    username   = Column(String(32), nullable=False, unique=True)
    first_name = Column(String(30))
    last_name  = Column(String(30))
    password   = Column(String(150))
    is_active  = Column(Boolean(create_constraint=False), nullable=False, default=False)
    date_joined = Column(DateTime, nullable=False, default=func.now())
    last_login  = Column(DateTime, nullable=True)
    roles = relationship("Role", secondary="auth_users_roles", lazy="joined")
    emails = relationship("AuthEmail", lazy="joined", back_populates="user")

    @property
    def is_admin(self):
        """Does user have a role called 'admin'?
        """
        for role in self.roles:
            if role.name == 'admin':
                return True
        return False
    
    @property
    def confirmed_emails(self):
        """Returns a list of confirmed emails.
        """
        return [e for e in self.emails if e.is_confirmed]
    
    @property
    def has_confirmed_emails(self):
        """Indicates whether the user has got any confirmed emails.
        """
        return bool(self.confirmed_emails)
    
    @property
    def preferred_email(self):
        if self.emails:
            preferred = [e for e in self.emails if e.is_preferred]
            if preferred:
                return preferred[0]
            return self.emails[0]
        return None
    
    @preferred_email.setter
    def set_preferred_email(self, email):
        for em in self.emails:
            if em is not email:
                em.is_preferred = False
        email.is_preferred = True
        if not email in self.emails:
            self.emails.append(email)
    
    def check_password(self, raw_password):
        if self.password != None:
            return _check_password(raw_password, self.password)
        return False

    def set_password(self, raw_password):
        if raw_password != None:
            self.password = _hash_password(raw_password)


class Role(Model, IdMixin):
    """A model for storing Role data.
    """
    __tablename__ = 'auth_roles'

    name = Column(String(30), nullable=False, unique=True)
    description = Column(String(256))


class AuthEmail(Model, IdMixin):
    """A model for storing Email addresses associated to a User model.
    """
    __tablename__ = 'auth_emails'

    address = Column(String(150), nullable=False, unique=True)
    user_id = Column(Integer, ForeignKey('auth_users.id'), nullable=False)
    confirmation_hash = Column(String(32), default=generate_confirmation_hash)
    is_confirmed = Column(Boolean(create_constraint=False), default=False)
    is_preferred = Column(Boolean(create_constraint=False), default=False)
    user = relationship("User", back_populates="emails")



# package: elixr_db.auth
import bcrypt
from datetime import datetime
from sqlalchemy.orm import exc as orm_exc
from .models import (
    _check_password, _hash_password,
    User, Role, AuthEmail
)


class Authenticator(object):
    """Encapsulates user authentication around an explicitly provided Session
    object. This seamlessly fits into whatever Session management strategy in
    use, and effectively `side-steps` whatever complexity that could arise from
    having to create a Session object locally.
    """
    def __init__(self, db_session, accept_email_as_username=False):
        if not db_session:
            raise ValueError('db_session is required.')
        
        self._db_session = db_session
        self._accept_email_as_username = accept_email_as_username
    
    def authenticate(self, username, password):
        db = self._db_session
        if self._accept_email_as_username and '@' in username:
            try:
                email = db.query(AuthEmail).filter_by(address=username).first()
                user = email.user if email and email.user and email.user.is_active else None
                if user and _check_password(password, user.password):
                    return user
                return None
            except orm_exc.NoResultFound:
                pass
            return None
        
        # if code gets this far it could be `accept_email_as_username=False` or
        # treatment of username as email failed thus need to treat username as 
        # just a username
        user = db.query(User).filter_by(username=username, is_active=True).first()
        if user and _check_password(password, user.password):
            return user
        return None
    
    def __call__(self, username, password):
        """Convenience method for calling `authenticate`.
        """
        return self.authenticate(username, password)
    


"""SqlAlchemy eXtension: a toolkit of reusable SQLAlchemy types, data models and 
utility functions encapsulating best practices or exposing convenient facades for
data access and handling.

Copyright (c) 2017, Hazeltek Solutions.
"""

__author__  = 'Hazeltek Solutions'
__version__ = '0.3'



from .meta import (
    BASE,
    Model,
)

from .types import (
    UUID,
    Choice
)

from .mixins import (
    IdMixin,
    TimestampMixin,
    IdTimestampMixin,
    UUIDMixin
)


"""Defines models which ease defining, storing and working with Addresses within
an application.
"""
from collections import namedtuple
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy import Column, Float, ForeignKey, Integer, String, \
        UniqueConstraint
from sqlalchemy.orm import relationship
from elixr.core import AttrDict, Coordinates
from .mixins import IdMixin
from .meta import Model



class Country(Model, IdMixin):
    """A model for storing Country data.
    """
    __tablename__ = 'countries'

    code = Column(String(3), nullable=False)
    name = Column(String(50), nullable=False, unique=True)
    states = relationship("State", back_populates="country")
    
    def __str__(self):
        return self.name


class State(Model, IdMixin):
    """A model for storing State data.
    """
    __tablename__ = 'states'
    __table_args__ = (
        UniqueConstraint('name', 'country_id', name='uq_states_name_country_id'),
    )

    code = Column(String(3), nullable=False)
    name = Column(String(50), nullable=False)
    country_id = Column(Integer, ForeignKey("countries.id"), nullable=False)
    country = relationship("Country", back_populates="states")

    def __str__(self):
        value = self.name
        if self.country:
            if value:
                value += ', '
            value += self.country.name
        return value


class AddressMixin(object):
    """A mixin which defines address fields for use within other models.
    """
    addr_raw    = Column(String(200))
    addr_street = Column(String(100))
    addr_town   = Column(String(50))
    addr_landmark = Column(String(100))
    postal_code = Column(String(10))
    
    @declared_attr
    def addr_state_id(cls):
        return Column(Integer, ForeignKey("states.id"), nullable=True)
        
    @declared_attr
    def addr_state(cls):
        return relationship("State")
        
    @property
    def address_dict(self):
        return self.to_dict(self)
    
    @property
    def address_str(self):
        return self.to_str(self)
    
    @staticmethod
    def to_dict(addr):
        addr_dict = dict(
            landmark = addr.addr_landmark,
            street = addr.addr_street,
            town = addr.addr_town,
            raw = addr.addr_raw,
        )
        if addr.addr_state:
            addr_dict['state'] = addr.addr_state.name
            addr_dict['state_code'] = addr.addr_state.code
            addr_dict['postal_code'] = addr.postal_code
            if addr.addr_state.country:
                addr_dict['country'] = addr.addr_state.country.name
                addr_dict['country_code'] = addr.addr_state.country.code
        return addr_dict

    @staticmethod
    def to_str(addr):
        value = ''
        if addr.addr_state:
            if addr.addr_street:
                value = addr.addr_street
            if addr.addr_town:
                if value:
                    value += ', '
                value += addr.addr_town
                if addr.postal_code:
                    value += ' %s' % addr.postal_code
            if value:
                value += ', '
            value += str(addr.addr_state)
            if addr.addr_landmark:
                if value:
                    value += ' '
                value += '(closest landmark: %s)' % addr.addr_landmark
        else:
            value = addr.addr_raw
        return value


class CoordinatesMixin(object):
    """A mixin which defines coordinate fields for use within other models.
    """
    longitude = Column(Float)
    latitude  = Column(Float)
    altitude  = Column(Float)
    gps_error = Column(Integer)

    @property
    def coordinates(self):
        return Coordinates(
            lng=self.longitude or 0.0, 
            lat=self.latitude or 0.0,
            alt=self.altitude, 
            error=self.gps_error)


class LocatableMixin(AddressMixin, CoordinatesMixin):
    """A convenience mixin which combines the fields defined by both the 
    AddressMixin and CoordinatesMixin.
    """
    pass


class Address(Model, IdMixin, CoordinatesMixin):
    """A model for storing Address data.
    """
    __tablename__ = 'addresses'
    
    raw    = Column(String(200), nullable=False)
    street = Column(String(100))
    town   = Column(String(50))
    landmark = Column(String(100))
    postal_code = Column(String(10))
    state_id = Column(Integer, ForeignKey("states.id"), nullable=True)
    state    = relationship("State", backref="addresses")

    def as_attrd(self):
        return AttrDict({
            'addr_raw': self.raw,
            'addr_street': self.street,
            'addr_town': self.town,
            'addr_state': self.state,
            'addr_landmark': self.landmark,
            'postal_code': self.postal_code,
        })

    def as_dict(self):
        return AddressMixin.to_dict(self.as_attrd())
    
    def __str__(self):
        return AddressMixin.to_str(self.as_attrd())



import sys
from datetime import date, datetime
from elixr.core import AttrDict
from ..address import Country, State
from ..orgz import PartyType, EmailContact, PhoneContact, Organisation


# python version flag
py3k = sys.version_info.major > 2


ERROR_NOT_INT = 'is invalid integer'
ERROR_NOT_DATE = 'has no date in it'
ERROR_NOT_FLOAT = 'is not a float'
ERROR_NOT_BOOLEAN = 'is not a boolean'
ERROR_MISSING_REQUIRED_TEXT = 'missing required text'
ERROR_NOT_UNICODE_OR_ASCII = 'not unicode or ascii string'
ERROR_NOT_ENUM_OF = lambda e: 'is bad integer/text for enum `%s`' % e.__name__

no_data = object()
no_enum = object()


def to_text(value):
    if isinstance(value, float):
        if int(value) == value:
            value = int(value)
    try:
        value = str(value)
    except UnicodeError:
        value = None
    return value


def to_bool(value):
    if value is not None:
        if type(value) == type(True):
            return value
        elif type(value) == type(0):
            return bool(value)

        # hack: needed for py2
        elif not py3k and 'long' in str(type(value)):
            return bool(value)
        
        value = to_text(value)
        if not value:
            return None
        if value.upper() in ['TRUE', 'YES', 'T', 'Y']:
            return True
        elif value.upper() in ['FALSE', 'NO', 'F', 'N']:
            return False
    return None


def to_enum(enum_type, value):
    if value is not None:
        if type(value) == type(0) or (not py3k and 'long' in str(type(value))):
            try:
                return enum_type(value)
            except ValueError:
                return None
        value = to_text(value)
        if value and hasattr(enum_type, value):
            return enum_type[value]
    return None


class XRefResolver(object):
    """Cross Refereence Resolver. Provides a convenient means of retrieving the
    `id` for a Model given a supposedly unique filter expression.

    It helps with cases where records are identified by a unique human-friendly
    value order than their respectively numeric id values. For performance, 
    results are cached for reuse.
    """
    def __init__(self, dbsession):
        self.__dbsession = dbsession
        self.__cache = {}
    
    def resolve(self, model_type, only_id=True, **filters):
        key = self.generate_key(model_type, only_id=only_id, **filters)
        if key not in self.__cache:
            fnquery = self.__dbsession.query
            query = fnquery(model_type.id if only_id else model_type) \
                        .filter_by(**filters)
            value = query.scalar() if only_id else query.one()
            if value:
                self.__cache[key] = value
        return self.__cache.get(key)
    
    def clear_cache(self):
        """Clears the cache maintained internally.
        """
        self.__cache.clear()
    
    @staticmethod
    def generate_key(model_type, only_id, **filters):
        prefix = model_type.__name__
        if only_id:
            prefix += '_id'
        pairs = ['%s=%s' % (k, filters[k]) for k in sorted(filters)]
        return ('#'.join([prefix] + pairs)).lower()


class ImporterBase(object):
    
    def __init__(self, context, progress_callback=None):
        assert 'db' in context
        assert 'cache' in context
        self.progress_callback = progress_callback
        if not isinstance(context, AttrDict):
            context = AttrDict(context)
        self.context = context
        self.errors = []
    
    def error(self, row, col, message):
        message_seq = (self.sheet_name, row, col, message)
        self.errors.append(message_seq)
    
    def resolve_xref(self, data, *xref_mapping):
        """Resolves Cross Reference records found within data using provided
        reference sources data.abs

        xref_mapping: is to be a list with items as thus:
            `(xref_field, filter_field, model_type)`
        """
        resolve = self.context.cache.resolve
        for xref_field, filter_field, model_type in xref_mapping:
            if xref_field not in data:
                continue
            only_id = xref_field.endswith('_id')
            xref_value = data.pop(xref_field)
            if not isinstance(xref_value, (list, tuple)):
                kw = {filter_field: xref_value}
                data[xref_field] = resolve(model_type, only_id, **kw)
            else:
                ids = []
                for value in xref_value:
                    kw = {filter_field: value}
                    ids.append(resolve(model_type, only_id, **kw))
                data[xref_field] = [id for id in ids if id]
        
    def progress(self, *args):
        if self.progress_callback is not None:
            self.progress_callback(*args)
    
    def is_empty_row(self, sheet, row, num_cols=30):
        # we've picked 30 as an arbitrary number of columns to test so that
        # caller doesn't specify the number.
        for col in range(1, num_cols):
            try:
                if sheet.cell(row=row, column=col).value:
                    return False
            except ValueError:
                break
        return True
    
    def get_cell_value(self, sheet, row, col, default=no_data):
        try:
            return sheet.cell(row=row, column=col).value
        except ValueError:
            if default != no_data:
                return default
            raise
    
    def get_cell_and_found(self, sheet, row, col, default=''):
        try:
            value = sheet.cell(row=row, column=col).value
            return (value, True) if value else (default, False)
        except ValueError:
            return (default, False)
    
    def get_bool_from_cell(self, sheet, row, col):
        value, found = self.get_cell_and_found(sheet, row, col)
        if not found:
            return None
        value = to_bool(value)
        if value is None:
            self.error(row, col, ERROR_NOT_BOOLEAN)
            return None
        return value
    
    def get_required_bool_from_cell(self, sheet, row, col):
        value, found = self.get_cell_and_found(sheet, row, col)
        if not value:
            self.error(row, col, ERROR_MISSING_REQUIRED_TEXT)
            return None
        return self.get_bool_from_cell(sheet, row, col)
    
    def get_date_from_cell(self, sheet, row, col, default=None):
        value, found = self.get_cell_and_found(sheet, row, col)
        if not found or value == '':
            return default
        
        if isinstance(value, datetime):
            return value.date()
        if isinstance(value, date):
            return value
        
        value = to_text(value)
        if value is not None:
            try:
                dt = [int(v) for v in value.replace('/', '-').split('-')]
                return date(*dt)
            except (TypeError, ValueError):
                self.error(row, col, ERROR_NOT_DATE)
        return None
    
    def get_required_date_from_cell(self, sheet, row, col):
        value, found = self.get_cell_and_found(sheet, row, col)
        if not found or value == '':
            self.error(row, col, ERROR_MISSING_REQUIRED_TEXT)
            return None
        return self.get_date_from_cell(sheet, row, col)
    
    def get_enum_from_cell(self, sheet, row, col, enum_type, default=no_enum):
        value, found = self.get_cell_and_found(sheet, row, col)
        if not found or value == '':
            if default is no_enum:
                self.error(row, col, ERROR_NOT_ENUM_OF(enum_type))
                return None
            return default
        value = to_enum(enum_type, value)
        if value is None:
            self.error(row, col, ERROR_NOT_ENUM_OF(enum_type))
        return value
    
    def get_required_enum_from_cell(self, sheet, row, col, enum_type):
        value, found = self.get_cell_and_found(sheet, row, col)
        if not found or value == '':
            self.error(row, col, ERROR_MISSING_REQUIRED_TEXT)
            return None
        return self.get_enum_from_cell(sheet, row, col, enum_type)

    def get_float_found_valid(self, sheet, row, col, default=0.0):
        value, found = self.get_cell_and_found(sheet, row, col, default)
        valid = True
        if isinstance(value, str) and not value.strip():
            value, found = (default, False)
        if found:
            try:
                value = float(value)
            except (TypeError, ValueError):
                self.error(row, col, ERROR_NOT_FLOAT)
                valid, value = (False, default or None)
        return (value, found, valid)
    
    def get_float_from_cell(self, sheet, row, col, default=0.0):
        value, found, valid = self.get_float_found_valid(sheet, row, col, default)
        return value
    
    def get_required_float_from_cell(self, sheet, row, col):
        value, found, valid = self.get_float_found_valid(sheet, row, col)
        if not found:
            self.error(row, col, ERROR_NOT_FLOAT)
        return value
    
    def get_id_from_cell(self, sheet, row, col, default=''):
        value, found, valid = self.get_text_found_valid(sheet, row, col, default)
        return value.strip()
    
    def get_required_id_from_cell(self, sheet, row, col):
        value, found, valid = self.get_text_found_valid(sheet, row, col)
        if valid and not value:
            self.error(row, col, ERROR_MISSING_REQUIRED_TEXT)
        return value.strip()
    
    def get_ids_from_cell(self, sheet, row, col):
        value, found, valid = self.get_text_found_valid(sheet, row, col)
        if not value:
            return []
        return [p.strip() for p in value.split(',') if p.strip()]
    
    def get_required_ids_from_cell(self, sheet, row, col):
        value, found = self.get_cell_and_found(sheet, row, col)
        if not value:
            self.error(row, col, ERROR_MISSING_REQUIRED_TEXT)
            return []
        return self.get_ids_from_cell(sheet, row, col)
    
    def get_int_found_valid(self, sheet, row, col, default=0):
        value, found = self.get_cell_and_found(sheet, row, col, default)
        valid = True
        if isinstance(value, str) and not value.strip():
            value, found = (default, False)
        if found:
            try:
                value = int(value)
            except (TypeError, ValueError):
                try:
                    value = float(value)
                except (TypeError, ValueError):
                    pass
            if isinstance(value, float):
                if int(value) != value:
                    self.error(row, col, ERROR_NOT_INT)
                    valid = False
                else:
                    value = int(value)
            elif not isinstance(value, int):
                self.error(row, col, ERROR_NOT_INT)
                valid, value = (False, None)
        return (value, found, valid)
    
    def get_int_from_cell(self, sheet, row, col, default=0):
        value, found, valid = self.get_int_found_valid(sheet, row, col, default)
        return value
    
    def get_required_int_from_cell(self, sheet, row, col):
        value, found, valid = self.get_int_found_valid(sheet, row, col)
        if not found:
            self.error(row, col, ERROR_NOT_INT)
        return value

    def get_text_found_valid(self, sheet, row, col, default=''):
        value, found = self.get_cell_and_found(sheet, row, col, default)
        valid = True
        if found:
            value = to_text(value)
            if not value:
                self.error(row, col, ERROR_NOT_UNICODE_OR_ASCII)
                valid, value = (False, default)
        return (value, found, valid)
    
    def get_text_from_cell(self, sheet, row, col, default=''):
        value, found, valid = self.get_text_found_valid(sheet, row, col, default)
        return value
    
    def get_required_text_from_cell(self, sheet, row, col):
        value, found, valid = self.get_text_found_valid(sheet, row, col)
        if valid and not value:
            self.error(row, col, ERROR_MISSING_REQUIRED_TEXT)
        return value

    @property
    def sheet(self):
        if self.sheet_name not in self.wb.sheetnames:
            return
        return self.wb.get_sheet_by_name(self.sheet_name)
    
    def import_data(self, wb):
        self.wb = wb
        if self.sheet:
            return self.process()


class ExactTextMatcher(object):
    """Returns all texts found in `available_texts` which are exact match 
    of `target_text`.
    """
    def __call__(self, available_texts, target_text):
        target_text = target_text.strip().lower()
        return [
            t for t in available_texts 
                if target_text == t.strip().lower()
        ]


class PrefixedTextMatcher(object):
    """Returns all texts found in `available_texts` which all start with
    `target_text` followed by some other text.
    """
    def __call__(self, available_texts, target_text):
        target_text = target_text.strip().lower()
        return [
            t for t in available_texts
                if t.strip().lower().startswith(target_text)
        ]


class SuffixedTextMatcher(object):
    """Returns all texts found in `available_texts` which all end with
    `target_text` and might have some other text in-front.abs
    """
    def __call__(self, available_texts, target_text):
        target_text = target_text.strip().lower()
        return [
            t for t in available_texts
                if t.strip().lower().endswith(target_text)
        ]


class MegaImporterBase(object):
    """Defines a base implementation for managing a collection of `importers`
    to pull in data from an excel file into a database.
    """
    def __init__(self, context, matcher=None):
        assert 'db' in context
        if not isinstance(context, AttrDict):
            context = AttrDict(context)
        self.matcher =  matcher or PrefixedTextMatcher()
        self.context = context
        self.errors = []
    
    @property
    def has_errors(self):
        return len(self.errors) > 0
    
    @property
    def importers(self):
        message = 'Needs to be overriden by derived classes.'
        raise NotImplementedError(message)
    
    def import_data(self, wb, progress_callback=None):
        matcher = self.matcher
        available_sheets = wb.get_sheet_names()
        for importer in self.importers:
            target_name = importer.sheet_name
            matches = matcher(available_sheets, target_name)
            for sheet_name in sorted(matches):
                importer.sheet_name = sheet_name
                imp = importer(self.context, progress_callback)
                imp.import_data(wb)
                self.errors.extend(imp.errors)
            
            # restore original sheet_name for importer
            importer.sheet_name = target_name
        
        if self.errors:
            self.context.db.rollback()
    
    def summarise_errors(self):
        errors = {}
        for sheet_name, row, col, message in self.errors:
            sheet_errors = errors.setdefault(sheet_name, {})
            sheet_errors.setdefault(message, []).append((col, row))
        
        error_lines = []
        for sheet_name, message_errors in sorted(errors.items()):
            if error_lines:
                error_lines.append('')
            error_lines.append(sheet_name)
            error_lines.append('-' * len(sheet_name))
            for message, cells in sorted(message_errors.items()):
                col_rows = []
                current_col, start, end = -1, 0, 0
                for col, row in sorted(cells):
                    if col != current_col or row > end + 1:
                        if current_col > -1:
                            col_rows.append((current_col, start, end))
                        current_col, start = col, row
                    end = row
                col_rows.append((current_col, start, end))
                error_lines.append('')
                error_lines.append(message + ':')

                error_cells = []
                for col, start, end in col_rows:
                    cell = chr(col + ord('A') - 1)
                    if start == end:
                        cell += '%s' % start
                    else:
                        cell += '%s-%s' % (start, end)
                    error_cells.append(cell)
                error_lines.append(', '.join(error_cells))
        return '\n'.join(error_lines)
    
    @classmethod
    def make_for_exact_match(cls, context):
        return cls(context, ExactTextMatcher())
    
    @classmethod
    def make_for_prefix_match(cls, context):
        return cls(context, PrefixedTextMatcher())
    
    @classmethod
    def make_for_suffix_match(cls, context):
        return cls(context, SuffixedTextMatcher())


class AdminBoundaryImporter(ImporterBase):
    """An importer able to proces and import Country and State data from an
    excel file.
    """
    sheet_name = 'admin-boundaries'

    def __init__(self, context, progress_callback=None):
        assert 'db' in context
        assert 'cache' in context
        super(AdminBoundaryImporter, self).__init__(context, progress_callback)
    
    def create_country(self, row, data):
        try:
            country = Country(**data)
            self.context.db.add(country)
        except Exception as ex:
            message_fmt = 'Country could not be created. Err: %s'
            self.error(row, 0, message_fmt % str(ex))
    
    def create_state(self, row, data):
        self.resolve_xref(data, ('country_id', 'code', Country))
        try:
            state = State(**data)
            self.context.db.add(state)
        except Exception as ex:
            message_fmt = 'State could not be created. Err: %s'
            self.error(row, 0, message_fmt % str(ex))
    
    def import_country(self, sh, row):
        row, nrows = (row + 2, sh.max_row)
        num_errors = len(self.errors)
        while row <= nrows:
            if self.is_empty_row(sh, row):
                break
            
            data = AttrDict()
            data['code'] = self.get_required_id_from_cell(sh, row, 1)
            data['name'] = self.get_required_text_from_cell(sh, row, 2)
            if num_errors < len(self.errors):
                break
            self.create_country(row, data)
            self.progress(row, nrows)
            row += 1
        return (row + 1)
    
    def import_state(self, sh, row):
        row, nrows = (row + 2, sh.max_row)
        num_errors = len(self.errors)
        while row <= nrows:
            if self.is_empty_row(sh, row):
                break
            
            data = AttrDict()
            data['code'] = self.get_required_id_from_cell(sh, row, 1)
            data['name'] = self.get_required_text_from_cell(sh, row, 2)
            data['country_id'] = self.get_required_id_from_cell(sh, row, 3)
            if num_errors < len(self.errors):
                break
            self.create_state(row, data)
            self.progress(row, nrows)
            row += 1
        return (row + 1)
        
    def process(self):
        sh, row, nrows = (self.sheet, 1, self.sheet.max_row)
        ## flags are provided in reverse order as `list.pop` operates
        ## on items from the bottom NOT the top
        flags = ['states', 'countries']
        while row <= nrows:
            if not flags: break
            if self.is_empty_row(sh, row):
                row += 1; continue
            
            value = sh.cell(row=row, column=1).value
            if value not in flags:
                row += 1; continue
            
            while value != flags.pop(): 
                pass # keep popping ;-)
            
            if value == 'countries':
                row = self.import_country(sh, row)
            elif value == 'states':
                row = self.import_state(sh, row)
        self.progress(row, nrows)


class PartyImporterBase(ImporterBase):
    """Base importer to pull excel data into a database for models derived
    from the Party model.
    """
    subtype = None

    def process_chunk(self, row, col, data):
        sh = self.sheet
        data['name'] = self.get_required_text_from_cell(sh, row, col)
        data['addr_street'] = self.get_text_from_cell(sh, row, col+1, None)
        data['addr_town'] = self.get_text_from_cell(sh, row, col+2, None)
        data['addr_state_id'] = self.get_required_id_from_cell(sh, row, col+3)
        data['addr_landmark'] = self.get_text_from_cell(sh, row, col+4, None)
        data['postal_code'] = self.get_text_from_cell(sh, row, col+5, None)
        return col+5
    
    def create_item(self, row, data):
        raise NotImplementedError()
    
    def add_item(self, row, item):
        raise NotImplementedError()
    
    def post_process(self):
        pass
    
    def process(self):
        sh, nrows = (self.sheet, self.sheet.max_row)
        for row in range(2, nrows + 1):
            data = AttrDict()
            num_errors = len(self.errors)
            self.process_chunk(row, 1, data)
            if num_errors == len(self.errors):
                item = self.create_item(row, data)
                self.add_item(row, item)
                self.progress(row, nrows)
        self.post_process()
        self.progress(row, nrows)


class OrganisationImporter(PartyImporterBase):
    """An importer to process and import Organisation data from an excel file.
    """
    IGNORE_REQ = '-x-'
    sheet_name = 'organisations'
    subtype = PartyType.organisation
    fncode_type = None

    def __init__(self, context, progress_callback=None):
        super(OrganisationImporter, self).__init__(context, progress_callback)
        self.__root = context.db.query(Organisation).first()
        assert self.fncode_type != None
    
    def add_item(self, row, item):
        if not item: return
        # be sure only single root exists
        if not item.parent and not item.parent_id:
            if self.__root:
                message = 'Root organisation already exists.'
                self.error(row, 0, message)
                return
            else:
                self.__root = item
        
        # all is good at this point
        self.context.db.add(item)
    
    def create_item(self, row, data):
        # extract contact fields
        contacts = []
        for email in (data.pop('emails') or []):
            contacts.append(EmailContact(address=email))
        for number in (data.pop('phones') or []):
            contacts.append(PhoneContact(number=number))
        
        self.resolve_xref(data, 
            ('addr_state_id', 'code', State),
            ('parent_id', 'short_name', Organisation))
        data.fncode = data.fncode.value     # norm into int
        try:
            item = Organisation(**data)
            if contacts:
                item.contacts.extend(contacts)
            return item
        except Exception as ex:
            message_fmt = 'Organisation could not be created. Err: %s'
            self.error(row, 0, message_fmt % str(ex))
        return None
    
    def post_process(self):
        pass
        
    def process_chunk(self, row, col, data):
        sh, ftype = (self.sheet, self.fncode_type)
        data['parent_id'] = self.get_required_id_from_cell(sh, row, col)
        data['identifier'] = self.get_required_id_from_cell(sh, row, col+1)
        data['fncode'] = self.get_required_enum_from_cell(sh, row, col+2, ftype)
        data['short_name'] = self.get_text_from_cell(sh, row, col+3, None)
        ## break-out to capture party chunk
        col = super(OrganisationImporter, self).process_chunk(row, col+4, data)
        ## return to normal flow
        data['website_url'] = self.get_text_from_cell(sh, row, col+1, None)
        data['emails'] = self.get_ids_from_cell(sh, row, col+2)
        data['phones'] = self.get_ids_from_cell(sh, row, col+3)
        data['date_established'] = self.get_date_from_cell(sh, row, col+4)
        data['description'] = self.get_text_from_cell(sh, row, col+5, None)
        data['longitude'] = self.get_float_from_cell(sh, row, col+6, None)
        data['latitude'] = self.get_float_from_cell(sh, row, col+7, None)
        data['altitude'] = self.get_float_from_cell(sh, row, col+8, None)
        data['gps_error'] = self.get_float_from_cell(sh, row, col+9, None)
        return col+12



import uuid
from sqlalchemy import Column, DateTime, Integer, Sequence, String
from sqlalchemy.sql import func
from .types import UUID



class IdMixin(object):
    """A mixin which defines an Id field to uniquely identify records within an
    application space.
    """
    id = Column(Integer, primary_key=True)


class TimestampMixin(object):
    """A mixin which defines fields to record timestamp for record creation and
    modification.
    """
    date_created = Column(DateTime, nullable=False, default=func.now())
    last_updated = Column(DateTime, nullable=True, onupdate=func.now())


class IdTimestampMixin(IdMixin, TimestampMixin):
    """A mixin which provides Id and timestamp fields for a records.
    """
    pass


class UUIDMixin(object):
    """A mixin which defines a globally unique identifier field for a record
    that can possibly be used to track an entity across applications.
    """
    uuid = Column(UUID, nullable=False, unique=True, default=uuid.uuid4)



import os
import elixr.sax
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md')) as f:
    README = f.read()
with open(os.path.join(here, 'CHANGES.md')) as f:
    CHANGES = f.read()


requires = [
    'sqlalchemy',
    
    # deps not on pypi
    'elixr.core'
]

tests_requires = [
    'bcrypt',
    'openpyxl',
    'pytest',
    'pytest-cov'
]

setup(
    name='elixr.sax',
    version=elixr.sax.__version__,
    description='Reusable types, models and utilities built around SQLAlchemy ORM',
    long_description=README + '\n\n' + CHANGES,
    author=elixr.sax.__author__,
    author_email='info@hazeltek.com',
    maintainer='Abdul-Hakeem Shaibu',
    maintainer_email='hkmshb@gmail.com',
    url='https://github.com/hkmshb/elixr.sax',
    keywords='elixr.sax, elixr SqlAlchemy eXtension',
    zip_safe=False,
    packages=find_packages(),
    platforms='any',
    install_requires=requires,
    extras_require={ 'testing': tests_requires },
    classifiers=[]
)


import uuid
from enum import Enum
from sqlalchemy.types import TypeDecorator, CHAR, Integer
from sqlalchemy.dialects.postgresql import UUID



class UUID(TypeDecorator):
    """Platform-independent UUID type.
    Uses PostgreSQL's UUID type, otherwise uses CHAR(32), storing as stringified
    hex values.

    # hint: adapted from sqlalchemy docs on 'Backend-agnostic GUID Type'
    """
    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))
    
    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int
    
    def process_result_value(self, value, dialect):
        if value is None:
            return value
        return uuid.UUID(value)


class Choice(TypeDecorator):
    """Choice offers way of having fixed set of choices for given column. It 
    works with :mod:`enum` in the standard library of Python 3.4+ (the enum34_
    backported package on PyPi is compatible too for ``< 3.4``).

    # hint: adapted from sqlalchemy_utils ChoiceType.
    """
    impl = Integer
    
    def __init__(self, enum_class, *args, **kwargs):
        super(Choice, self).__init__(*args, **kwargs)
        self.enum_class = enum_class
    
    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        return self.enum_class(value).value
    
    def process_result_value(self, value, dialect):
        if value is None:
            return value
        return self.enum_class(value)



"""Defines models which ease defining, storing and working with Organisational
structures within an application.
"""
import enum
from sqlalchemy import Column, Boolean, Date, ForeignKey, Integer, String, Table
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship, backref

from .mixins import IdMixin, IdTimestampMixin, UUIDMixin
from .address import AddressMixin, CoordinatesMixin
from .types import Choice
from .meta import Model



## ENUMS
class ContactType(enum.Enum):
    email = 1
    phone = 2


class PartyType(enum.Enum):
    person = 1
    organisation = 2


class Gender(enum.Enum):
    unknown = 0
    male    = 1
    female  = 2


class MaritalStatus(enum.Enum):
    unknown  = 0
    single   = 1
    married  = 2
    divorced = 3
    widowed  = 4



## MODELS
# many-to-many relation between contact details and party
parties_contact_details = Table(
    'parties_contact_details',
    Model.metadata,
    Column('party_id', Integer, ForeignKey('parties.id')),
    Column('contact_detail_id', Integer, ForeignKey('contact_details.id'))
)


class ContactDetail(Model, IdMixin):
    """A Single-Table Inheritance model for storing all forms of contact details.
    """
    __tablename__ = 'contact_details'
    
    usage = Column(String(30))
    subtype = Column(Choice(ContactType), nullable=False)
    is_confirmed = Column(Boolean(create_constraint=False), default=False)
    is_preferred = Column(Boolean(create_constraint=False), default=False)
    deleted = Column(Boolean(create_constraint=False), default=False)
    __mapper_args__ = {
        'polymorphic_identity': 'contact_details',
        'polymorphic_on': subtype
    }

class EmailContact(ContactDetail):
    """A model for storing Email contact details.
    """
    __mapper_args__ = {
        'polymorphic_identity': ContactType.email
    }
    address = Column(String(150), unique=True)


class PhoneContact(ContactDetail):
    """A model for storing Phone contact details.
    """
    __mapper_args__ = {
        'polymorphic_identity': ContactType.phone
    }
    number = Column(String(15), unique=True)
    extension = Column(String(10))


class Party(Model, IdTimestampMixin, AddressMixin):
    """A Joined Table inheritance model for storing the named parts of a Party
    derived inheritance model relationship.
    """
    __tablename__ = 'parties'

    name = Column(String(50), nullable=False, unique=True)
    subtype = Column(Choice(PartyType), nullable=False)
    deleted = Column(Boolean(create_constraint=False), default=False)
    contacts = relationship("ContactDetail", secondary="parties_contact_details", 
                            lazy="joined")
    __mapper_args__ = {
        'polymorphic_identity': 'parties',
        'polymorphic_on': subtype
    }

class Person(Party):
    """A model for storing Person details.
    """
    __tablename__ = 'people'
    __mapper_args__ = {
        'polymorphic_identity': PartyType.person
    }

    id = Column(Integer, ForeignKey("parties.id"), primary_key=True)
    title = Column(String(20))
    middle_name = Column(String(50))
    last_name = Column(String(50))
    gender = Column(Choice(Gender))
    date_born = Column(Date)
    marital_status = Column(Choice(MaritalStatus))
    state_origin_id = Column(Integer, ForeignKey("states.id"))
    state_origin = relationship("State")
    nationality_id  = Column(Integer, ForeignKey("countries.id"))
    nationality = relationship("Country")

    @property
    def first_name(self):
        return self.name
    
    @first_name.setter
    def set_first_name(self, value):
        self.name = value


class Organisation(Party, UUIDMixin, CoordinatesMixin):
    """A model for storing Organisation details.
    
    :hint: fncode (function code) can be used to indicate the organisation 
    function implemented via an enum. e.g HQ, Branch, SalesOffice, ServicePoint
    etc. Its left open to accomodate any int based enum declaration.
    """
    __tablename__ = 'organisations'
    __mapper_args__ = {
        'polymorphic_identity': PartyType.organisation
    }

    id = Column(Integer, ForeignKey("parties.id"), primary_key=True)
    parent_id = Column(Integer, ForeignKey("organisations.id"))
    fncode = Column(Integer)                                         # function code
    identifier = Column(String(30), nullable=False, unique=True)     # org identifier
    short_name = Column(String(15), unique=True)
    description = Column(String(255))
    date_established = Column(Date)
    website_url = Column(String(150))
    children = relationship("Organisation", foreign_keys=[parent_id],
                backref=backref("parent", remote_side=[id]))



from collections import namedtuple
from .meta import Model, create_engine, sessionmaker



Resource = namedtuple('Resource', ['engine', 'session'])

def make_session(conn_str='sqlite:///:memory:', initdb_callback=None):
    engine = create_engine(conn_str)
    Model.metadata.create_all(engine)

    SASession = sessionmaker(bind=engine)
    session = SASession()

    if initdb_callback:
        initdb_callback(session)
    return Resource(engine, session)


def clear_tables(db, *table_names):
    tables = reversed(Model.metadata.sorted_tables)
    if table_names:
        tables = [t for t in tables if t.name in table_names]

    for table in tables:
        db.execute(table.delete())
    db.commit()


def drop_tables(engine):
    Model.metadata.drop_all(engine)
